<?php

/**
 * Handles plate/image push from Weizmann CRIMS to IceBear.
 */
class WisCrimsImport extends Device {

    const USERNAME='wiscrims';
    const USERFULLNAME='WIS CRIMS';

    /**
     * @param $params
     * @return array[]
     * @throws BadRequestException
     * @throws ForbiddenException
     * @throws NotFoundException
     * @throws ServerException
     */
    public static function createImages($params){
        static::require('images',$params);
        $createdImages=[];
        foreach ($params['images'] as $image){
            $createdImages[]=static::createImage($image);
        }
        return array('created'=>$createdImages);
    }

    /**
     * Creates an image record, or updates an existing one, with the supplied UUencoded image and thumbnail.
     * @param $image
     * @return array|array[]
     * @throws BadRequestException
     * @throws ServerException
     * @throws NotFoundException
     * @throws ForbiddenException
     */
    public static function createImage($image){
        if(isset($image['images'])){
            return static::createImages($image);
        }
        static::require('plateBarcode',$image);
        static::require('rowNumber',$image);
        static::require('columnNumber',$image);
        static::require('dropNumber',$image);
        static::require('image',$image);
        static::require('thumbnail',$image);
        session::becomeAdmin();
        $barcode=$image['plateBarcode'];
        $row=1*$image['rowNumber'];
        $col=1*$image['columnNumber'];
        $sub=1*$image['dropNumber'];
        if(!$row){
            $rowLabels=platetype::$rowLabels;
            $row=array_search($image['rowNumber'], $rowLabels);
        }
        $limsPlate=plate::getByName($barcode);
        if(!$limsPlate){
            throw new NotFoundException('No plate with barcode '.$barcode);
        }
        $wellDrops=plate::getwelldrops($limsPlate['id']);
        $limsPlate['welldrops']=$wellDrops['rows'];

        $imageName=platewell::getWellLabel($row, $col).'_'.$sub.'.tmp';
        $dummyInspection=static::getDummyPlateInspectionForPlate($limsPlate);
        $inspectionId=$dummyInspection['id'];
        foreach($limsPlate['welldrops'] as $drop){
            if($drop['row']==$row && $drop['col']==$col && $drop['dropnumber']==$sub){
                $wellDropId=$drop['id'];
                break;
            }
        }
        if(!isset($wellDropId)){
            throw new NotFoundException("Plate $barcode has no drop at row $row, column $col, drop $sub");
        }
        $inspectionImage=dropimage::getByProperties(array(
            'welldropid'=>$wellDropId,
            'imagingsessionid'=>$inspectionId
        ));
        if(isset($inspectionImage['rows']) && 1==count($inspectionImage['rows'])){
            $inspectionImage=$inspectionImage['rows'][0];
        } else {
            $inspectionImage=static::createDummyImage($inspectionId, $wellDropId);
        }
        if(!$inspectionImage){
            throw new NotFoundException("No image with inspection ID $inspectionId and welldrop ID $wellDropId. Could not create dummy image.");
        }

        $inspectionImageId=$inspectionImage['id'];

        $imageStore=rtrim(config::get('core_imagestore'),'/').'/';
        $imagePath=substr($barcode,0,2).'/'.substr($barcode,0,3).'/'.$barcode.'/';
        $imagePath=$imageStore.$imagePath.'imagingsession'.$inspectionId.'/';
        $thumbPath=$imagePath.'thumbs/';
        @unlink($imagePath.platewell::getWellLabel($row, $col).'_'.$sub.'.jpg');
        @unlink($thumbPath.platewell::getWellLabel($row, $col).'_'.$sub.'.jpg');

        try {
            if(!file_exists($thumbPath) && !@mkdir($thumbPath, 0755, true)){
                throw new ServerException('Could not create directory for image storage');
            }

            $imagePath.=$imageName;
            $thumbPath.=$imageName;

            if(!@file_put_contents($imagePath, convert_uudecode($image['image'])) ){
                throw new ServerException('Could not write image to image store');
            }
            if(!@file_put_contents($thumbPath, convert_uudecode($image['thumbnail'])) ){
                throw new ServerException('Could not write thumbnail to image store');
            }
            if(strlen(@file_get_contents($imagePath))<10){
                //probably a 2-byte garbage file caused by UUencode header/footer
                $imageBytes=static::stripUUencodingHeaderAndFooterLines($image['image']);
                $imageBytes=convert_uudecode($imageBytes);
                if(!@file_put_contents($imagePath, $imageBytes) ){
                    throw new ServerException('Could not write image to image store');
                }
            }
            if(strlen(@file_get_contents($thumbPath))<10){
                //probably a 2-byte garbage file caused by UUencode header/footer
                $thumbBytes=static::stripUUencodingHeaderAndFooterLines($image['thumbnail']);
                $thumbBytes=convert_uudecode($thumbBytes);
                if(!@file_put_contents($thumbPath, $thumbBytes) ){
                    throw new ServerException('Could not write thumbnail to image store');
                }
            }

            $imageProperties=@getimagesize($imagePath);
            $thumbProperties=@getimagesize($thumbPath);
            if(!$imageProperties || !$imageProperties[0] || !$thumbProperties || !$thumbProperties[0]){
                throw new ServerException('Could not determine size of image and/or thumbnail. Is it corrupt?');
            }
            if(IMAGETYPE_JPEG==$imageProperties[2]){
                $newImagePath=substr($imagePath, 0, -4).'.jpg';
            } else if(IMAGETYPE_PNG==$imageProperties[2]){
                $newImagePath=substr($imagePath, 0, -4).'.png';
            } else {
                throw new ServerException('Image must be PNG or JPEG');
            }
            if(IMAGETYPE_JPEG==$thumbProperties[2]){
                $newThumbPath=substr($thumbPath, 0, -4).'.jpg';
            } else if(IMAGETYPE_PNG==$thumbProperties[2]){
                $newThumbPath=substr($thumbPath, 0, -4).'.png';
            } else {
                throw new ServerException('Thumbnail must be PNG or JPEG');
            }
            if(!@rename($imagePath, $newImagePath) || !@rename($thumbPath, $newThumbPath)){
                throw new ServerException('Could not save image/thumb in final store');
            }
            $width=$imageProperties[0];
            $height=$imageProperties[1];

        } catch (ServerException $e){
            @unlink($imagePath);
            @unlink($thumbPath);
            throw $e;
        }

        dropimage::update($inspectionImageId, array(
            'pixelheight'=>$height,
            'pixelwidth'=>$width,
//            'micronsperpixelx'=>$micronsPerPixel,
//            'micronsperpixely'=>$micronsPerPixel
            'imagepath'=>str_replace($imageStore,'',$newImagePath),
            'thumbnailpath'=>str_replace($imageStore,'',$newThumbPath),
        ));

        $limsImage=dropimage::getById($inspectionImageId);
        session::revertAdmin();
        return $limsImage;
    }

    private static function stripUUencodingHeaderAndFooterLines($rawString){
        $parts=explode("\n", $rawString, 2);
        $headerlessString=$parts[1];
        $parts=explode("\nend", $headerlessString);
        $cleanString=$parts[0];
        return $cleanString;
    }

    /**
     * Creates multiple plates.
     * @param $params
     * @return array[]
     * @throws BadRequestException
     * @throws ForbiddenException
     * @throws NotFoundException
     * @throws NotModifiedException
     * @throws ServerException
     */
    public static function createPlates($params){
        if(isset($params['plateType'])){ return static::createPlate($params); }
        static::require('plates',$params);
        $createdPlates=[];
        foreach($params['plates'] as $plate){
            $createdPlates[]=static::createPlate($plate);
        }
        return array('created'=>$createdPlates);
    }

    /**
     * Creates a plate.
     * @param $plate array The plate - requires scoringSystemName, plateType, barcode, owner
     * @return array
     * @throws BadRequestException
     * @throws ForbiddenException
     * @throws NotFoundException
     * @throws ServerException
     * @throws NotModifiedException
     */
    public static function createPlate($plate){
        session::becomeAdmin();
        if(isset($plate['plates'])){ return static::createPlates($plate); }
        static::require('scoringSystemName',$plate);
        static::require('plateType',$plate);
        static::require('barcode',$plate);
        static::require('owner',$plate);
        if('CRIMS'!==$plate['scoringSystemName']){
            throw new BadRequestException('Plate scoringSystemName must be "CRIMS"');
        }

        $barcode=$plate['barcode'];
        $defaultProjectId=static::getDefaultProjectId();
        $plateType=static::createPlateType($plate['plateType']);
        $scoringSystem=static::getScoringSystem();
        $owner=static::createUser($plate['owner']);

        $limsPlate=plate::getByName($plate['barcode']);
        if(!$limsPlate){
            $limsPlate=plate::create(array(
                'name'=>$barcode,
                'platetypeid'=>$plateType['id'],
                'crystalscoringsystemid'=>$scoringSystem['id'],
                'ownerid'=>$owner['id'],
                'projectid'=>$defaultProjectId
            ));
            $limsPlate=$limsPlate['created'];
        }
        $wellDrops=plate::getwelldrops($limsPlate['id']);
        $limsPlate['welldrops']=$wellDrops['rows'];

        $limsInspection=static::getDummyPlateInspectionForPlate($limsPlate);

        if(isset($plate['proteinConstruct'])) {
            static::createProteinAndConstruct($limsPlate['id'], $plate['proteinConstruct'], $owner);
        }

        if(isset($plate['interestingDrops'])){
            foreach ($plate['interestingDrops'] as $drop){
                static::handleInterestingDrop($drop, $limsPlate, $limsInspection);
            }
        }

        session::revertAdmin();
        unset($limsPlate['welldrops']);
        return $limsPlate;
    }

    /**
     * Creates project, protein and construct if needed, and sets the construct on the plate.
     * @param $limsPlateId
     * @param $proteinConstruct
     * @param $owner
     * @throws BadRequestException
     * @throws ForbiddenException
     * @throws NotFoundException
     * @throws NotModifiedException
     * @throws ServerException
     */
    public static function createProteinAndConstruct($limsPlateId, $proteinConstruct, $owner){
        static::require('proteinName',$proteinConstruct);
        if(!isset($proteinConstruct['proteinAcronym'])) {
            $proteinConstruct['proteinAcronym']='';
        }
        if(!isset($proteinConstruct['constructName'])){
            $proteinConstruct['constructName']='Construct 1';
        }
        $proteinName=$proteinConstruct['proteinName'];
        $constructName=$proteinConstruct['constructName'];

        try {
            $project=project::getByName($proteinName);
        } catch (NotFoundException $e){
            $project=project::create(array(
                'name'=>$proteinName,
                'description'=>$proteinName,
                'owner'=>session::getUserId()
            ));
            $project=$project['created'];
        }
        $projectId=$project['id'];

        $proteins=project::getproteins($projectId);
        if(!empty($proteins)){
            foreach ($proteins['rows'] as $p){
                if($p['name']==$proteinName){
                    $protein=$p;
                    break;
                }
            }
        }
        if(!isset($protein)){
            $protein=protein::create(array(
                'name'=>$proteinName,
                'description'=>$proteinName,
                'projectid'=>$projectId,
                'proteinacronym'=>$proteinConstruct['proteinAcronym']
            ))['created'];
        }
        $proteinId=$protein['id'];

        $constructs=protein::getconstructs($proteinId);
        foreach ($constructs['rows'] as $c){
            if($c['name']==$constructName){
                $construct=$c;
                break;
            }
        }
        if(!isset($construct)){
            $construct=construct::create(array(
                'proteinid'=>$proteinId,
                'projectid'=>$projectId,
                'name'=>$constructName,
                'description'=>$constructName
            ))['created'];
            $sequences=[];
            if(isset($proteinConstruct['sequences'])){
                $sequences=$proteinConstruct['sequences'];
            } else if(isset($proteinConstruct['dnaSequences'])) {
                $sequences=$proteinConstruct['dnaSequences'];
            }
            if(!empty($sequences)){
                $count=1;
                foreach($sequences as $sequence){
                    $dnaSequence='';
                    $proteinSequence='';
                    $sequence=preg_replace('[^A-Z]','',strtoupper($sequence));
                    if(preg_match('[^ACGT]',$sequence)){
                        //it's a protein sequence
                        $proteinSequence=$sequence;
                    } else {
                        //it's a DNA sequence
                        $dnaSequence=$sequence;
                        $converted=sequence::dnaToProtein($sequence);
                        if($converted){
                            $proteinSequence=$converted;
                        }
                    }
                    sequence::create(array(
                        'constructid'=>$construct['id'],
                        'projectid'=>$projectId,
                        'name'=>'Sequence '.$count,
                        'description'=>'Sequence '.$count,
                        'dnasequence'=>$dnaSequence,
                        'proteinsequence'=>$proteinSequence
                    ));
                    $count++;
                }
            }
        }
        $constructId=$construct['id'];

        plate::update($limsPlateId, array(
            'constructid'=>$constructId
        ));

        project::update($project['id'],array(
            'owner'=>$owner['id']
        ));

    }

    /**
     * Creates the user, or returns the existing one.
     * @param $user array The user to create.
     * @return array The user - requires name, username, emailAddress
     * @throws BadRequestException
     * @throws ForbiddenException
     * @throws NotFoundException
     * @throws ServerException
     */
    public static function createUser($user){
        static::require('name',$user);
        static::require('username',$user);
        static::require('emailAddress',$user);
        $username=$user['username'];
        $fullName=$user['name'];
        $email=$user['emailAddress'];
        $limsUser=user::getByName($username);
        if($limsUser){
            if($limsUser['email']!=$email){
                throw new BadRequestException('User '.$username.' email address mismatch. Source: '.$email.', IceBear: '.$limsUser['email']);
            }
            return $limsUser;
        }
        $limsUser=user::create(array(
            'name'=>$username,
            'fullname'=>$fullName,
            'email'=>$email,
            'password'=>'USELESSPASSWORD',
            'isactive'=>0
        ));
        return $limsUser['created'];
    }

    /**
     * Creates the plate type, or returns the existing one
     * @param $plateType array The plate type - requires name,rows,columns,subPositions
     * @return array
     * @throws BadRequestException
     * @throws ForbiddenException
     * @throws NotFoundException
     * @throws ServerException
     */
    public static function createPlateType($plateType){
        static::require('name',$plateType);
        static::require('rows',$plateType);
        static::require('columns',$plateType);
        static::require('subPositions',$plateType);
        $name=$plateType['name'];
        $rows=1*$plateType['rows'];
        $cols=1*$plateType['columns'];
        $drops=1*$plateType['subPositions'];
        if($rows<=0 || $cols<=0 || $drops <=0){
            throw new BadRequestException('Bad plate type geometry - rows, columns, subPositions must all be integers above zero');
        }
        $limsPlateType=platetype::getByName($name);
        if($limsPlateType){
            $limsLayout=$limsPlateType['rows'].'x'.$limsPlateType['cols'].'x'.$limsPlateType['subs'];
            $sourceLayout=$rows.'x'.$cols.'x'.$drops;
            if($limsLayout!=$sourceLayout){
                throw new BadRequestException('Plate type '.$name.' layout mismatch. Source: '.$sourceLayout.', destination '.$limsLayout);
            }
        } else {
            $sharedProjectId=static::getSharedProjectId();
            $defaultDropMapping='';
            $reservoir='';
            for($i=1;$i<=(int)$drops;$i++){
                $defaultDropMapping.=$i;
                $reservoir.='R';
            }
            $defaultDropMapping.=','.$reservoir; //e.g., 123,RRR
            $limsPlateType=platetype::create(array(
                'name'=>$name,
                'rows'=>$rows,
                'cols'=>$cols,
                'subs'=>$drops,
                'dropmapping'=>$defaultDropMapping,
                'projectid'=>$sharedProjectId
            ));
            $limsPlateType=$limsPlateType['created'];
        }
        return $limsPlateType;
    }

    /**
     * Returns the CRIMS scoring system, creating it and its scores if not found.
     * @return array The scoring system
     * @throws BadRequestException
     * @throws ForbiddenException
     * @throws NotFoundException
     * @throws ServerException
     */
    private static function getScoringSystem(){
        $crimsSS=crystalscoringsystem::getByName('CRIMS');
        if(!empty($crimsSS)){
            return $crimsSS;
        }
        $sharedProjectId=static::getSharedProjectId();
        $crimsSS=crystalscoringsystem::create(array(
            'projectid'=>$sharedProjectId,
            'name'=>'CRIMS',
            'iscurrent'=>1
        ));
        $crimsSS=$crimsSS['created'];
        $scores=array(
            array('index'=>0, 'color'=>'7fff00', 'name'=>'Clear drop'),
            array('index'=>1, 'color'=>'0e1ef5', 'name'=>'Heavy Precipitate'),
            array('index'=>2, 'color'=>'7ab9d4', 'name'=>'Light Precipitate'),
            array('index'=>8, 'color'=>'8b4513', 'name'=>'No Diffraction'),
            array('index'=>9, 'color'=>'8b4513', 'name'=>'Low Resolution'),
            array('index'=>7, 'color'=>'8b4513', 'name'=>'Salt')
        );
        foreach($scores as $s){
            crystalscore::create(array(
                'projectid'=>$sharedProjectId,
                'crystalscoringsystemid'=>$crimsSS['id'],
                'label'=>$s['name'],
                'name'=>'CRIMS_'.$s['name'],
                'color'=>$s['color'],
                'scoreindex'=>$s['index'],
                'hotkey'=>$s['index']
            ));
        }
        return $crimsSS;
    }

    /**
     * Returns a dummy inspection for the plate, creating it if necessary.
     * @param $plate array The plate - requires name, id
     * @return array|mixed
     * @throws BadRequestException
     * @throws ForbiddenException
     * @throws NotFoundException
     * @throws ServerException
     */
    private static function getDummyPlateInspectionForPlate($plate){
        $plateBarcode=$plate['name'];
        $dummyInspection=imagingsession::getByName($plateBarcode.'_dummy');
        if(!$dummyInspection){
            $dummyImager=imager::getByName('+20 Microscope');
            if(!$dummyImager){
                throw new NotFoundException('No imager in IceBear called +20 Microscope');
            }
            $dummyInspection=imagingsession::create(array(
                'name'=>$plateBarcode.'_dummy',
                'manufacturerdatabaseid'=>0,
                'plateid'=>$plate['id'],
                'imagerid'=>$dummyImager['id'],
                'imageddatetime'=>gmdate('Y-m-d H:i:s')
            ));
            $dummyInspection=$dummyInspection['created'];
        }
        return $dummyInspection;
    }

    /**
     * @param $drop
     * @param $limsPlate
     * @param $limsInspection
     * @throws BadRequestException
     * @throws ForbiddenException
     * @throws NotFoundException
     * @throws ServerException
     */
    private static function handleInterestingDrop($drop, $limsPlate, $limsInspection){
        static::require('rowNumber',$drop);
        static::require('columnNumber',$drop);
        static::require('dropNumber',$drop);
        $row=1*$drop['rowNumber'];
        $col=1*$drop['columnNumber'];
        $sub=1*$drop['dropNumber'];
        if(!$row){
            $rowLabels=platetype::$rowLabels;
            $row=array_search($drop['rowNumber'], $rowLabels);
        }
        foreach($limsPlate['welldrops'] as $welldrop){
            if($row==$welldrop['row'] && $col==$welldrop['col'] && $sub==$welldrop['dropnumber']){
                $limsDrop=$welldrop;
            }
        }
        if(!isset($limsDrop)){
            throw new NotFoundException("handleInterestingDrop: No drop with row $row, col $col, sub $sub");
        }
        $dropId=$limsDrop['id'];
        $inspectionId=$limsInspection['id'];
        $limsDropImage=static::createDummyImage($inspectionId, $dropId);
        if(isset($drop['crystals'])){
            foreach ($drop['crystals'] as $crystal){
                static::handleCrystal($crystal, $limsDropImage);
            }
        }
    }

    /**
     * @param $limsInspectionId
     * @param $wellDropId
     * @return array|mixed
     * @throws BadRequestException
     * @throws ForbiddenException
     * @throws NotFoundException
     * @throws ServerException
     */
    private static function createDummyImage($limsInspectionId, $wellDropId){
        $wellDrop=welldrop::getById($wellDropId);
        if(!$wellDropId){ throw new NotFoundException("No well drop with ID $wellDropId"); }
        $plateWell=platewell::getById($wellDrop['platewellid']);
        $plate=plate::getById($plateWell['plateid']);
        $barcode=$plate['name'];
        $row=$plateWell['row'];
        $col=$plateWell['col'];
        $sub=$wellDrop['dropnumber'];
        $imageStore=rtrim(config::get('core_imagestore'),'/').'/';
        $imagePath='dummyimage.jpg';
        $thumbPath='dummythumb.jpg';
        $dummyFilename=platewell::getWellLabel($row, $col).'_'.$sub.'.jpg';
        $midPath=substr($barcode, 0, 2).'/'.substr($barcode,0,3).'/'.$barcode.'/imagingsession'.$limsInspectionId.'/';
        if(!file_exists($imageStore.$midPath.'thumbs') && !@mkdir($imageStore.$midPath.'thumbs', 0755, true)){
            throw new ServerException('Could not create image/thumbnail path: '.$imageStore.$midPath.'thumbs');
        }
        if(!@copy($imageStore.$imagePath, $imageStore.$midPath.$dummyFilename)){
            throw new ServerException('Could not copy dummy image into place');
        }
        if(!@copy($imageStore.$thumbPath, $imageStore.$midPath.'thumbs/'.$dummyFilename)){
            throw new ServerException('Could not copy dummy image into place');
        }

        $imageName=$plate['name'].'_'.platewell::getWellLabel($row, $col).'.'.$sub.'_is'.$limsInspectionId;
        $limsDropImage=dropimage::getByName($imageName);
        if(!$limsDropImage){
            $limsDropImage=dropimage::create(array(
                'name'=>$imageName,
                'projectid'=>$wellDrop['projectid'],
                'imagingsessionid'=>$limsInspectionId,
                'welldropid'=>$wellDropId,
                'imagestorepath'=>$imageStore,
                'imagepath'=>$imagePath,
                'thumbnailpath'=>$thumbPath
            ))['created'];
        }
        return $limsDropImage;
    }

    /**
     * @param $crystal
     * @param $limsDropImage
     * @return array
     * @throws BadRequestException
     * @throws ForbiddenException
     * @throws NotFoundException
     * @throws ServerException
     */
    private static function handleCrystal($crystal, $limsDropImage){
        static::require('sampleName',$crystal);
        static::require('numberInDrop',$crystal);
        static::require('sendersImage',$crystal);
        $image=$crystal['sendersImage'];
        static::require('micronsPerPixel',$image);
        static::require('marks',$image);
        if(1!==count($image['marks'])){ throw new BadRequestException('Exactly one mark per crystal'); }
        $mark=$image['marks'][0];
        if('point'!==$mark['regionType']){ throw new BadRequestException('Only point marks are supported'); }
        static::require('x',$mark);
        static::require('y',$mark);
        $pixelX=$mark['x'];
        $pixelY=$mark['y'];
        $limsDropImageId=$limsDropImage['id'];
        $limsWellDrop=welldrop::getbyId($limsDropImage['welldropid']);
        $existing=crystal::getByProperties(array(
            'dropimageid'=>$limsDropImageId,
            'pixelx'=>$pixelX,
            'pixely'=>$pixelY
        ));
        if(!empty($existing)){ return $existing['rows'][0]; }
        $numCrystalsInDrop=0;
        $crystalsInDrop=welldrop::getcrystals($limsDropImage['welldropid']);
        if(isset($crystalsInDrop['rows'])){
            $numCrystalsInDrop=count($crystalsInDrop['rows']);
        }
        $numberInDrop=1+$numCrystalsInDrop;
        $crystalName=str_replace('.','d',$limsWellDrop['name']).'c'.$numberInDrop;

        $fields=array(
            'prefix'=>$crystalName,
            'name'=>$crystalName,
            'numberindrop'=>$numberInDrop,
            'dropimageid'=>$limsDropImageId,
            'welldropid'=>$limsWellDrop['id'],
            'projectid'=>$limsWellDrop['projectid'],
            'pixelx'=>$pixelX,
            'pixely'=>$pixelY
        );
        foreach(array("spaceGroup","unitCellA","unitCellB","unitCellC","unitCellAlpha","unitCellBeta","unitCellGamma") as $field){
            if(isset($crystal[$field]) && !empty ($crystal[$field])){
                $fields[strtolower($field)]=$crystal[$field];
            }
        }
        $limsCrystal=crystal::create($fields);

        if(isset($image['micronsperpixel'])){
            dropimage::update($limsDropImageId, array(
                'micronsperpixelx'=>$image['micronsperpixel'],
                'micronsperpixely'=>$image['micronsperpixel'],
            ));
        }
        return $limsCrystal;
    }

}
